# -*- coding: utf-8 -*-

"""
Figure Manager
==============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib.backend_bases import FigureManagerBase


# All declaration
__all__ = ['FigureManager']


# %% CLASS DEFINITIONS
# Custom FigureManager class
class FigureManager(FigureManagerBase):
    pass
