# -*- coding utf-8 -*-

"""
Utils
=====
Contains various utility functions and non-widget classes for *GuiPy*.

"""


# %% IMPORTS
# Import base modules

# All declaration
__all__ = []

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"
