# -*- coding: utf-8 -*-

"""
Data Table Selection Model
==========================

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports

# All declaration
__all__ = ['DataTableSelectionModel']


# %% CLASS DEFINITIONS
# Define model for the DataTable widget
# HINT: https://doc.qt.io/qt-5/model-view-programming.html
class DataTableSelectionModel(QC.QItemSelectionModel):
    pass
